// Example code for homework 1.
// You do not have to use this code if you don't want, but it may be helpful
// to get you started.

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// helpful typedefs
typedef void (*InitialConditions)(float *x, float *c, float *radius);
typedef void (*VelocityField)(const float x[3], float t, float v[3]);
typedef void (*VelocityGradient)(const float x[3], float t, float dvdx[3][3]);
typedef void (*ParticleIntegrator)(float *x, float t, float dt);

// function prototypes
void process_args(int argc, const char **argv);
void initialize_particles(float **x, float **c, float **radius);
void free_particles(float **x, float **c, float **radius);
void output_particles(int frame, float *x, float *c, float *radius);
void solve_3x3(float A[3][3], const float b[3], float x[3]); // A gets overwritten

// velocity field etc.
void given_conditions(float *x, float *c, float *radius);
void given_velocity(const float x[3], float t, float v[3]);
void given_gradient(const float x[3], float t, float dvdx[3][3]);
void custom_conditions(float *x, float *c, float *radius);
void custom_velocity(const float x[3], float t, float v[3]);
void custom_gradient(const float x[3], float t, float dvdx[3][3]);

// integrators
void forward_euler(float *x, float t, float dt);
void backward_euler(float *x, float t, float dt);
void trapezoidal_rule(float *x, float t, float dt);
void runge_kutta3(float *x, float t, float dt);

// global parameters
InitialConditions initial_conditions;
VelocityField velocity_field;
VelocityGradient velocity_gradient;
int n;
int end_seconds;
int frames_per_second;
ParticleIntegrator particle_integrator;
const char *output_directory;

// function definitions
int main(int argc, const char **argv)
{
   float *x, *c, *radius;
   float t=0, dt;
   int frame=0;

   process_args(argc, argv);
   initialize_particles(&x, &c, &radius);
   output_particles(frame, x, c, radius);
   while(frame<end_seconds*frames_per_second){
      t=frame/(float)frames_per_second;
      dt=1.0/frames_per_second;
      particle_integrator(x, t, dt);
      ++frame;
      output_particles(frame, x, c, radius);
      // give some feedback that we're doing something
      fputc(frame%10 ? '+' : '#', stdout); fflush(stdout);
   }
   free_particles(&x, &c, &radius);
   return 0;
}

void process_args(int argc, const char **argv)
{
   int i;

   // defaults
   initial_conditions=given_conditions;
   velocity_field=given_velocity;
   velocity_gradient=given_gradient;
   n=1000;
   end_seconds=10;
   particle_integrator=forward_euler;
   frames_per_second=24;
   output_directory="."; // i.e. just dump data in the current directory
   // look to see if these were overridden
   for(i=1; i<argc; ++i){
      if(strcmp(argv[i], "-v")==0){
         if(i+1<argc){
            ++i;
            switch(argv[i][0]){
               case '0':
                  initial_conditions=given_conditions;
                  velocity_field=given_velocity;
                  velocity_gradient=given_gradient;
                  break;
               case '1':
                  initial_conditions=custom_conditions;
                  velocity_field=custom_velocity;
                  velocity_gradient=custom_gradient;
                  break;
            }
         }
      }else if(strcmp(argv[i], "-n")==0){
         if(i+1<argc){
            ++i;
            sscanf(argv[i], "%d", &n);
         }
      }else if(strcmp(argv[i], "-t")==0){
         if(i+1<argc){
            ++i;
            sscanf(argv[i], "%d", &end_seconds);
         }
      }else if(strcmp(argv[i], "-f")==0){
         if(i+1<argc){
            ++i;
            sscanf(argv[i], "%d", &frames_per_second);
         }
      }else if(strcmp(argv[i], "-m")==0){
         if(i+1<argc){
            ++i;
            switch(argv[i][0]){
               case 'f': case 'F':
                  particle_integrator=forward_euler;
                  break;
               case 'b': case 'B':
                  particle_integrator=backward_euler;
                  break;
               case 't': case 'T':
                  particle_integrator=trapezoidal_rule;
                  break;
               case 'r': case 'R':
                  particle_integrator=runge_kutta3;
                  break;
            }
         }
      }else if(strcmp(argv[i], "-o")==0){
         if(i+1<argc){
            ++i;
            output_directory=argv[i];
         }
      }else{
         printf("unknown command-line option %s\n", argv[i]);
      }
   }
}

void initialize_particles(float **x, float **c, float **radius)
{
   *x=calloc(n, 3*sizeof(float));
   *c=calloc(n, 3*sizeof(float));
   *radius=calloc(n, sizeof(float));
   initial_conditions(*x, *c, *radius);
}

void free_particles(float **x, float **c, float **radius)
{
   free(*x);
   *x=0;
   free(*c);
   *c=0;
   free(*radius);
   *radius=0;
}

void output_particles(int frame, float *x, float *c, float *radius)
{
   char *file_name=malloc(strlen(output_directory)+20);
   FILE *fp;
   int i;

   // the output format is ASCII:
   //  an integer giving the number of particles, followed by that many lines containing
   //  the position (3 floats) and colour (3 floats) of each particle.

   sprintf(file_name, "%s/particles%04d", output_directory, frame);
   fp=fopen(file_name, "wt");
   if(!fp){
      fprintf(stderr,"Could not open %s for writing\n", file_name);
      exit(1);
   }
   free(file_name);
   fprintf(fp, "%d\n", n);
   for(i=0; i<n; ++i){
      fprintf(fp, "%.7g %.7g %.7g %.7g %.7g %.7g %.7g\n", x[3*i], x[3*i+1], x[3*i+2], c[3*i], c[3*i+1], c[3*i+2], radius[i]);
   }
   fclose(fp);
}

void given_conditions(float *x, float *c, float *radius)
{
   // @@@@@@@@@@@@
   // Fill in code here to randomly distribute particles in the cylinder specified in the assignment.
   // Also assign particle colours to something pretty.
}

void given_velocity(const float x[3], float t, float v[3])
{
   v[0]=-25*x[2]*x[1];
   v[1]=0;
   v[2]=25*x[0]*x[1];
}

void given_gradient(const float x[3], float t, float dvdx[3][3])
{
   // @@@@@@@@@@@@
   // Fill in code here to calculate the gradient of the velocity field.
}

void custom_conditions(float *x, float *c, float *radius)
{
   // @@@@@@@@@@@@
   // Fill in custom initial conditions for particles.
}

void custom_velocity(const float x[3], float t, float v[3])
{
   // @@@@@@@@@@@@
   // Fill in custom velocity field for particles.
}

void custom_gradient(const float x[3], float t, float dvdx[3][3])
{
   // @@@@@@@@@@@@
   // Fill in gradient of custom velocity field for particles.
}

void solve_3x3(float A[3][3], const float b[3], float x[3])
{
   // Could use LU with pivoting for better stability, but chances are that if stability
   // is an issue for a 3x3 system, either the Newton solver stands no chance of converging,
   // or even if it does, the numerical solution fo the differential equation will be
   // garbage anyways. (Note: it's probably too slow to use determinant formulas past size 3x3,
   // even if the matrix is well-conditioned enough; use LAPACK instead)
   float minor00=A[1][1]*A[2][2]-A[1][2]*A[2][1],
         minor10=A[2][1]*A[0][2]-A[2][2]*A[0][1],
         minor20=A[0][1]*A[1][2]-A[0][2]*A[1][1];
   float over_det=1.0f/(A[0][0]*minor00+A[1][0]*minor10+A[2][0]*minor20);
   x[0]=(b[0]*minor00+b[1]*minor10+b[2]*minor20)*over_det;
   x[1]=(b[0]*(A[1][2]*A[2][0]-A[1][0]*A[2][2])
        +b[1]*(A[2][2]*A[0][0]-A[2][0]*A[0][2])
        +b[2]*(A[0][2]*A[1][0]-A[0][0]*A[1][2]))*over_det;
   x[2]=(b[0]*(A[1][0]*A[2][1]-A[1][1]*A[2][0])
        +b[1]*(A[2][0]*A[0][1]-A[2][1]*A[0][0])
        +b[2]*(A[0][0]*A[1][1]-A[0][1]*A[1][0]))*over_det;
}

void forward_euler(float *x, float t, float dt)
{
   // @@@@@@@@@@@@
   // Fill in code to advance particles (x) at time t by time step dt with Forward Euler.
}

void backward_euler(float *x, float t, float dt)
{
   // @@@@@@@@@@@@
   // Fill in code to advance particles (x) at time t by time step dt with Backward Euler.
}

void trapezoidal_rule(float *x, float t, float dt)
{
   // @@@@@@@@@@@@
   // Fill in code to advance particles (x) at time t by time step dt with Trapezoidal Rule.
}

void runge_kutta3(float *x, float t, float dt)
{
   // @@@@@@@@@@@@
   // Fill in code to advance particles (x) at time t by time step dt with TVD Runge-Kutta 3.
}

