#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// globals for rendering
int fullscreen_mode=0;
int width=800, height=600;
int saved_width, saved_height;
int oldmousex=0, oldmousey=0;
enum {DONOTHING, ROTATE, ZOOM, MOVETARGET} mousemode=DONOTHING;
float camdist=4;
const float NEARCLIP_FACTOR=0.0625, FARCLIP_FACTOR=4;
float nearclip, farclip;
float target[3]={0, 1, 0};
float view_target_timer=0;
int view_ground=0;
float old_time=0;
const float fovy=45;
float heading=M_PI/3, pitch=0;
float campos[3];
int file_count=0;

// globals for data
int frame=0;
const char *directory=0;
int n=0, allocn=0;
float *px=0;
float *pc=0;

// prototypes
void read_frame(int new_frame);
void handle_display(void);
void handle_reshape(int w, int h);
void handle_keypress(unsigned char key, int x, int y);
void handle_click(int button, int state, int x, int y);
void handle_drag(int x, int y);
void handle_idle(void);
void toggle_fullscreen(void);
void compute_camera_position(void);
void display_ground(void);
void display_particles(void);
void display_target(void);
void save_screen(const char *filename);

// definitions
int main(int argc, char *argv[])
{
   // initialize GLUT stuff 
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
   glutInitWindowSize(width,height);
   saved_width=width;
   saved_height=height;
   glutCreateWindow("hw1_viewer");
   glutDisplayFunc(handle_display);
   glutReshapeFunc(handle_reshape);
   glutKeyboardFunc(handle_keypress);
   glutMouseFunc(handle_click);
   glutMotionFunc(handle_drag);
   glutIdleFunc(handle_idle);

   // initialize OpenGL stuff
   glEnable(GL_DEPTH_TEST);
   glClearColor(0, 0, 0, 0);
   glClearDepth(1);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   nearclip=NEARCLIP_FACTOR*camdist;
   farclip=FARCLIP_FACTOR*camdist;

   if(argc<2){
      fprintf(stderr, "Expecting at least one argument (path to directory containing output)\n");
      return 1;
   }
   directory=argv[1];

   // initialize first couple of frames
   read_frame(0);

   // let's go
   old_time=clock()/(float)CLOCKS_PER_SEC;
   glutPostRedisplay();
   glutMainLoop();

   free(px);
   free(pc);
   return 0;
}

void read_frame(int new_frame)
{
   char *file_name;
   FILE *fp;
   int i;
   float radius;

   if(new_frame<0) return;
   file_name=malloc(strlen(directory)+20);
   sprintf(file_name, "%s/particles%04d", directory, new_frame);
   fp=fopen(file_name, "rt");
   if(fp){
      int new_n;
      frame=new_frame;
      fscanf(fp, "%d", &new_n);
      if(px==0){
         px=malloc(3*new_n*sizeof(float));
         pc=malloc(3*new_n*sizeof(float));
         allocn=new_n;
      }else if(new_n>allocn){
         new_n=allocn; // really should expand arrays to account for this instead of truncating!
      }
      n=new_n;
      for(i=0; i<n; ++i){
         fscanf(fp, "%f %f %f %f %f %f %f", px+3*i, px+3*i+1, px+3*i+2, pc+3*i, pc+3*i+1, pc+3*i+2, &radius);
         // note: for now, we ignore the radius
         // if you want to see the radius, you'll have to extend this viewer a little
      }
      fclose(fp);
   }
}

void handle_display(void)
{
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(campos[0], campos[1], campos[2], target[0], target[1], target[2], 0, 1, 0);

   if (view_ground) display_ground();
   display_particles();
   if(view_target_timer>0) display_target();

   glutSwapBuffers();
}

void handle_reshape(int w, int h)
{
   width=w;
   height=h;
   glViewport(0, 0, (GLsizei)width, (GLsizei)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(fovy, width/(float)height, nearclip, farclip);
   glutPostRedisplay();
}

void handle_keypress(unsigned char key, int x, int y)
{
   switch(key){
   case '.':
      read_frame(frame+1);
      break;
   case ',':
      read_frame(frame-1);
      break;
   case 'f':
      toggle_fullscreen();
      break;
   case 'g':
      view_ground=!view_ground;
      break;
   case 'q':
      exit(0);
   case 's':
      {
         char file_name[30];
         sprintf(file_name, "hw1_%04d.ppm", file_count);
         save_screen(file_name);
         ++file_count;
      }
      break;
   default:
      return;
   }
   glutPostRedisplay();
}

void handle_click(int button, int state, int x, int y)
{
   int mods=glutGetModifiers();
   if(state==GLUT_UP) mousemode = DONOTHING;
   else if(mods&GLUT_ACTIVE_SHIFT){
      view_target_timer=1;
      mousemode=MOVETARGET;
   }else{
      if(button==GLUT_RIGHT_BUTTON || (button==GLUT_LEFT_BUTTON && mods&GLUT_ACTIVE_CTRL))
         mousemode=ZOOM;
      else if(button==GLUT_LEFT_BUTTON)
         mousemode=ROTATE;
   }
   oldmousex=x;
   oldmousey=y;
}

void handle_drag(int x, int y)
{
   switch(mousemode){
   case ROTATE:
      heading+=0.007*(oldmousex-x);
      if(heading<-M_PI) heading+=2*M_PI;
      else if(heading>M_PI) heading-=2*M_PI;
      pitch+=0.007*(oldmousey-y);
      if(pitch<-.48*M_PI) pitch=-.48*M_PI;
      else if(pitch>.48*M_PI) pitch=.48*M_PI;
      break;
   case ZOOM:
      camdist*=pow(1.01, y-oldmousey);
      nearclip=NEARCLIP_FACTOR*camdist;
      farclip=FARCLIP_FACTOR*camdist;
      break;
   case MOVETARGET:
      target[0]-=(0.002*camdist)*cos(heading)*(oldmousex-x);
      target[1]-=(0.002*camdist)*(oldmousey-y);
      target[2]-=(0.002*camdist)*sin(heading)*(oldmousex-x);
      break;
   case DONOTHING:
      break;
   }
   oldmousex=x;
   oldmousey=y;
   compute_camera_position();
   glutPostRedisplay();
}

void handle_idle(void)
{
   float new_time=clock()/(float)CLOCKS_PER_SEC;
   float delta_time=new_time-old_time;
   int target_turned_off=0;
   if(delta_time<0) delta_time=0; // this can happen due to roll-over
   old_time=new_time;
   if(view_target_timer>0){
      view_target_timer-=delta_time;
      if(view_target_timer<=0){
         target_turned_off=1;
         view_target_timer=0;
      }
   }
   if(target_turned_off)
      glutPostRedisplay();
}

void toggle_fullscreen(void)
{
   if(fullscreen_mode){
      fullscreen_mode=0;
      width=saved_width;
      height=saved_height;
      glutReshapeWindow(saved_width, saved_height);
   }else{
      fullscreen_mode=1;
      saved_width=width;
      saved_height=height;
      glutFullScreen();
   }
}

void compute_camera_position(void)
{
   float ch=cos(heading), sh=sin(heading), cp=cos(pitch), sp=sin(pitch);
   campos[0]=target[0]-camdist*sh*cp;
   campos[1]=target[1]-camdist*sp;
   campos[2]=target[2]-camdist*ch*cp;
}

void display_ground(void)
{
   glBegin(GL_QUADS);
   glColor3f(.3, .3, .3);
   glVertex3f(-25, 0, -25);
   glVertex3f(25, 0, -25);
   glVertex3f(25, 0, 25);
   glVertex3f(-25, 0, 25);
   glEnd();
}

void display_particles(void)
{
   int i;
   glBegin(GL_POINTS);
   for(i=0; i<n; ++i){
      glColor3fv(pc+3*i);
      glVertex3fv(px+3*i);
   }
   glEnd();
}

void display_target(void)
{
   float logcd=log(camdist)/log(20.);
   float smallsize=pow(20.0, floor(logcd-.5));
   glMatrixMode (GL_MODELVIEW);
   glPushMatrix();
   glTranslatef(target[0], target[1], target[2]);
   glColor3f(1, 1, 1);
   glutWireCube(smallsize);
   glutWireCube(20*smallsize);
   glPopMatrix();
}

void save_screen(const char *filename)
{
   FILE *fp;
   GLubyte *image_buffer;

   // get the pixels
   image_buffer=malloc(3*width*height*sizeof(GLubyte));
   glReadBuffer(GL_FRONT);
   glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, image_buffer);
  
   // write them to the file
   fp=fopen(filename, "wb");
   if(fp){
      int i;
      fprintf(fp, "P6\n%d %d 255\n", width, height);
      for(i=1; i<=height; ++i){
         fwrite(image_buffer+3*width*(height-i), 1, 3*width, fp);
      }
   }
   
   // clean up
   free(image_buffer);
}

